### =========================================================================
### The transcripts(), exons() and cds() extractors
### -------------------------------------------------------------------------


### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### A high-level representation of the db relational model used for
### generating SQL queries.
###

### Note that we omit the *_start and *_end cols.
.ALLCOLS <- c("gene_id",
              "tx_id", "tx_name", "tx_chrom", "tx_strand",
              "exon_id", "exon_name", "exon_chrom", "exon_strand",
              "cds_id", "cds_name", "cds_chrom", "cds_strand",
              "exon_rank")

.CORETAGS <- c("id", "chrom", "strand", "start", "end")

### THE TRANSCRIPT CENTRIC POINT OF VIEW:
### If we look at the db from a transcript centric point of view, then the
### graph of relations between the tables becomes a tree where the 'transcript'
### table is the root:
###                            transcript
###                             /      \
###                          gene    splicing
###                                  /      \
###                                exon     cds
###
.TRANSCRIPT_CENTRIC_DBDESC <- list(
    CORECOLS=structure(paste("tx", .CORETAGS, sep="_"), names=.CORETAGS),
    ### Each col defined in the db is assigned to the closest table (starting
    ### from the 'transcript' table) where it can be found. The 'transcript'
    ### element must be the 1st in the list:
    COLMAP=list(
        transcript=c("tx_id", makeFeatureColnames("tx")),
        gene="gene_id",
        splicing=c("exon_rank", "exon_id", "_exon_id", "cds_id", "_cds_id"),
        exon=c("exon_name", "exon_chrom", "exon_strand",
               "exon_start", "exon_end"),
        cds=c("cds_name", "cds_chrom", "cds_strand", "cds_start", "cds_end")
    ),
    ### For each table that is not the root or a leaf table in the above tree,
    ### we list the tables that are below it:
    CHILDTABLES=list(
        splicing=c("exon", "cds")
    ),
    ### For each table that is not the root table in the above tree, we specify
    ### the join condition with the parent table:
    JOINS=c(
        gene="transcript._tx_id=gene._tx_id",
        splicing="transcript._tx_id=splicing._tx_id",
        exon="splicing._exon_id=exon._exon_id",
        cds="splicing._cds_id=cds._cds_id"
    )
)

### THE EXON CENTRIC POINT OF VIEW:
### If we look at the db from an exon centric point of view, then the
### graph of relations between the tables becomes a tree where the 'exon'
### table is the root:
###                               exon
###                                |
###                             splicing
###                            /   |    \
###                   transcript  gene  cds
###
.EXON_CENTRIC_DBDESC <- list(
    CORECOLS=structure(paste("exon", .CORETAGS, sep="_"), names=.CORETAGS),
    ### Each col defined in the db is assigned to the closest table (starting
    ### from the 'exon' table) where it can be found. The 'exon' element must
    ### be the 1st in the list:
    COLMAP=list(
        exon=c("exon_id", makeFeatureColnames("exon")),
        splicing=c("tx_id", "_tx_id", "exon_rank", "cds_id", "_cds_id"),
        transcript=c("tx_name", "tx_chrom", "tx_strand", "tx_start", "tx_end"),
        gene="gene_id",
        cds=c("cds_name", "cds_chrom", "cds_strand", "cds_start", "cds_end")
    ),
    ### For each table that is not the root or a leaf table in the above tree,
    ### we list the tables that are below it:
    CHILDTABLES=list(
        splicing=c("transcript", "gene", "cds")
    ),
    ### For each table that is not the root table in the above tree, we specify
    ### the join condition with the parent table:
    JOINS=c(
        splicing="exon._exon_id=splicing._exon_id",
        transcript="splicing._tx_id=transcript._tx_id",
        gene="splicing._tx_id=gene._tx_id",
        cds="splicing._cds_id=cds._cds_id"
    )
)

### THE CDS CENTRIC POINT OF VIEW:
### If we look at the db from a cds centric point of view, then the
### graph of relations between the tables becomes a tree where the 'cds'
### table is the root:
###                               cds
###                                |
###                             splicing
###                            /   |    \
###                   transcript  gene  exon
###
.CDS_CENTRIC_DBDESC <- list(
    CORECOLS=structure(paste("cds", .CORETAGS, sep="_"), names=.CORETAGS),
    ### Each col defined in the db is assigned to the closest table (starting
    ### from the 'cds' table) where it can be found. The 'cds`' element must
    ### be the 1st in the list:
    COLMAP=list(
        cds=c("cds_id", makeFeatureColnames("cds")),
        splicing=c("tx_id", "_tx_id", "exon_rank", "exon_id", "_exon_id"),
        transcript=c("tx_name", "tx_chrom", "tx_strand", "tx_start", "tx_end"),
        gene="gene_id",
        exon=c("exon_name", "exon_chrom", "exon_strand",
               "exon_start", "exon_end")
    ),
    ### For each table that is not the root or a leaf table in the above tree,
    ### we list the tables that are below it:
    CHILDTABLES=list(
        splicing=c("transcript", "gene", "exon")
    ),
    ### For each table that is not the root table in the above tree, we specify
    ### the join condition with the parent table:
    JOINS=c(
        splicing="cds._cds_id=splicing._cds_id",
        transcript="splicing._tx_id=transcript._tx_id",
        gene="splicing._tx_id=gene._tx_id",
        exon="splicing._exon_id=exon._exon_id"
    )
)

.DBDESC <- list(
    transcript=.TRANSCRIPT_CENTRIC_DBDESC,
    exon=.EXON_CENTRIC_DBDESC,
    cds=.CDS_CENTRIC_DBDESC
)


### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Some utility functions.
###

.checkargColumns <- function(columns, valid_columns)
{
    if (is.null(columns))
        return()
    if (is.character(columns) && all(columns %in% valid_columns))
        return()
    valid_columns <- paste("\"", valid_columns, "\"", sep="", collapse = ", ")
    stop("'columns' must be NULL or a character vector ",
         "with values in ", valid_columns)
}

.getClosestTable <- function(root_table, colnames)
{
    COLMAP <- .DBDESC[[root_table]]$COLMAP
    ans <- character(length(colnames))
    ans[] <- NA_character_
    for (tablename in names(COLMAP))
        ans[colnames %in% COLMAP[[tablename]]] <- tablename
    ans
}

.assignColToClosestTable <- function(root_table, colnames)
{
    COLMAP <- .DBDESC[[root_table]]$COLMAP
    lapply(COLMAP, function(cols) intersect(cols, colnames))
}

.asQualifiedColnames <- function(root_table, colnames)
{
    colnames[colnames == "tx_id"] <- "_tx_id"
    colnames[colnames == "exon_id"] <- "_exon_id"
    colnames[colnames == "cds_id"] <- "_cds_id"
    paste(.getClosestTable(root_table, colnames), colnames, sep=".")
}

.joinRootToChildTables <- function(root_table, child_tables)
{
    COLMAP <- .DBDESC[[root_table]]$COLMAP
    CHILDTABLES <- .DBDESC[[root_table]]$CHILDTABLES
    JOINS <- .DBDESC[[root_table]]$JOINS
    all_tables <- names(COLMAP)
    ans <- all_tables[1L]
    for (i in seq_len(length(all_tables))[-1L]) {
        right_table <- all_tables[i]
        right_children <- c(right_table, CHILDTABLES[[right_table]])
        if (length(intersect(child_tables, right_children)) != 0L)
            ans <- paste(ans, "LEFT JOIN", right_table,
                              "ON", JOINS[[right_table]])
    }
    ans
}

### In the case of TranscriptDb objects, the distance between 'root_table'
### and 'child_table' is always <= 2.
### TODO: Revisit this. Not sure it would be guaranteed to work correctly if
### the distance between 'root_table' and 'child_table' was >= 3.
.joinPrimaryKeyToChildTable <- function(root_table, child_table)
{
    COLMAP <- .DBDESC[[root_table]]$COLMAP
    CHILDTABLES <- .DBDESC[[root_table]]$CHILDTABLES
    JOINS <- .DBDESC[[root_table]]$JOINS
    all_tables <- names(COLMAP)
    ans <- ""
    for (i in seq_len(length(all_tables))[-1L]) {
        right_table <- all_tables[i]
        right_children <- c(right_table, CHILDTABLES[[right_table]])
        if (length(intersect(child_table, right_children)) != 0L) {
            if (ans == "") {
                ans <- right_table
                next
            }
            ans <- paste(ans, "LEFT JOIN", right_table,
                              "ON", JOINS[[right_table]])
        }
    }
    ans
}

### convert a named list into an SQL where condition
.sqlWhereIn <- function(vals)
{
    if (length(vals) == 0L)
        return("")
    sql <-
      lapply(seq_len(length(vals)), function(i) {
               v <- vals[[i]]
               if (!is.numeric(v))
                 v <- paste("'", v, "'", sep="")
               v <- paste("(", paste(v, collapse=","), ")", sep="")
               v <- paste(names(vals)[i], " IN ", v, sep="")
               paste("(", v, ")", sep="")
            })
    paste("WHERE", paste(unlist(sql), collapse = " AND "))
}

.extractData <- function(root_table, txdb, what_cols, child_tables, vals,
                         orderby_cols=NULL)
{
    SQL_what <- paste(.asQualifiedColnames(root_table, what_cols),
                      collapse=", ")
    SQL_from <- .joinRootToChildTables(root_table, child_tables)
    SQL_where <- .sqlWhereIn(vals)
    if (length(orderby_cols) == 0L)
        SQL_orderby <- ""
    else
        SQL_orderby <- paste("ORDER BY", paste(orderby_cols, collapse=", "))
    SQL <- paste("SELECT DISTINCT", SQL_what, "FROM", SQL_from,
                 SQL_where, SQL_orderby)
    ans <- dbEasyQuery(txdbConn(txdb), SQL)
    names(ans) <- what_cols
    ans
}

.getWhereCols <- function(vals)
{
    if (is.null(vals))
        return(character(0))
    ans <- NULL
    if (is.list(vals)) {
        if (length(vals) == 0L)
            return(character(0))
        ans <- names(vals)
    }
    if (is.null(ans))
        stop("'vals' must be NULL or a named list")
    #valid_columns <- setdiff(.ALLCOLS, "exon_rank")
    valid_columns <- .ALLCOLS
    if (!all(ans %in% valid_columns)) {
        valid_columns <- paste("\"", valid_columns, "\"",
                               sep="", collapse = ", ")
        stop("'vals' must be NULL or a list with names ",
             "in ", valid_columns)
    }
    ans
}

.extractRootData <- function(root_table, txdb, vals, root_columns)
{
    CORECOLS <- .DBDESC[[root_table]]$CORECOLS
    orderby_cols <- CORECOLS[c("chrom", "strand", "start", "end")]
    what_cols <- unique(c(CORECOLS, root_columns))
    where_cols <- .getWhereCols(vals)
    if (is.list(vals))
        names(vals) <- .asQualifiedColnames(root_table, where_cols)
    where_tables <- unique(.getClosestTable(root_table, where_cols))
    .extractData(root_table, txdb, what_cols, where_tables, vals, orderby_cols)
}

.extractDataFromChildTable <- function(root_table, txdb,
                                       primary_key, ids,
                                       child_table, child_columns)
{
    ans_names <- c(primary_key, child_columns)
    primary_key <- paste("_", primary_key, sep="")
    what_cols <- c(primary_key,
                   .asQualifiedColnames(root_table, child_columns))
    SQL_what <- paste(what_cols, collapse=", ")
    SQL_from <- .joinPrimaryKeyToChildTable(root_table, child_table)
    vals <- list(ids)
    names(vals) <- primary_key
    SQL_where <- .sqlWhereIn(vals)
    SQL <- paste("SELECT DISTINCT", SQL_what, "FROM", SQL_from, SQL_where)
    ans <- dbEasyQuery(txdbConn(txdb), SQL)
    names(ans) <- ans_names
    ans
}

.extractChildData <- function(root_table, txdb, ids, assigned_columns)
{
    primary_key <- .DBDESC[[root_table]]$CORECOLS["id"]
    ans <- NULL
    all_tables <- names(assigned_columns)
    for (i in seq_len(length(all_tables))[-1L]) {
        child_columns <- assigned_columns[[i]]
        if (length(child_columns) == 0L)
            next
        child_table <- all_tables[i]
        data0 <- .extractDataFromChildTable(root_table, txdb,
                                            primary_key, ids,
                                            child_table, child_columns)
        data <- lapply(data0[ , -1L, drop=FALSE],
                       function(col0)
                       {
                           col <- split(col0, data0[[1L]])
                           col <- col[as.character(ids)]
                           class0 <- class(col0)
                           class <- paste(toupper(substr(class0, 1L, 1L)),
                                          substr(class0, 2L, nchar(class0)),
                                          "List", sep="")
                           get(class)(unname(col))
                       })
        data <- DataFrame(data)
        if (is.null(ans))
            ans <- data
        else
            ans <- c(ans, data)
    }
    ans
}

.extractFeatureRowsAsGRanges <- function(root_table, txdb, vals, columns)
{
    CORECOLS <- .DBDESC[[root_table]]$CORECOLS
    forbidden_columns <- c(CORECOLS["chrom"], CORECOLS["strand"])
    .checkargColumns(columns, setdiff(.ALLCOLS, forbidden_columns))
    assigned_columns <- .assignColToClosestTable(root_table, columns)
    root_columns <- assigned_columns[[root_table]]
    ## Extract the data from the db.
    root_data <- .extractRootData(root_table, txdb, vals, root_columns)
    child_data <- .extractChildData(root_table, txdb,
                          root_data[[CORECOLS["id"]]], assigned_columns)
    ## Construct the GRanges object and return it.
    ## TODO: Use seqinfo <- seqinfo(txdb) when this becomes available.
    ans_seqinfo <- getTranscriptDbSeqinfo(txdb)
    ans_seqnames <- factor(root_data[[CORECOLS["chrom"]]],
                           levels=seqnames(ans_seqinfo))
    ans_ranges <- IRanges(start=root_data[[CORECOLS["start"]]],
                          end=root_data[[CORECOLS["end"]]])
    ans_strand <- strand(root_data[[CORECOLS["strand"]]])
    ans <- GRanges(seqnames=ans_seqnames,
                   ranges=ans_ranges,
                   strand=ans_strand)
    ## TODO: Use seqinfo(ans) <- ans_seqinfo when this becomes available.
    ans@seqinfo <- ans_seqinfo
    ans_values <- c(DataFrame(root_data[root_columns]), child_data)
    values(ans) <- ans_values[columns]
    ans
}


### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### The top-level extractors.
###
### Note that the current naming of the args is a little bit confusing
### because we have the 'vals' and 'columns' args and it is the latter that is
### related to the "values" slot of the returned GRanges object, not the
### former. More precisely, the user specifies what goes into the "values"
### slot (aka "elementMetadata" slot) thru the 'columns' arg and not thru the
### 'vals' arg.
### TODO:
###   - Rename the 'vals' arg -> 'filter'.
###   - Rename the 'columns' arg -> 'colnames'.
###

setGeneric("transcripts", function(x, ...) standardGeneric("transcripts"))

setMethod("transcripts", "data.frame",
    function(x, vals=NULL, columns=c("tx_id", "tx_name"))
        stop("Please use 'transcripts_deprecated' for older ",
             "data.frame-based transcript metadata.")
)

setMethod("transcripts", "TranscriptDb",
    function(x, vals=NULL, columns=c("tx_id", "tx_name"))
        .extractFeatureRowsAsGRanges("transcript", x, vals, columns)
)

setGeneric("exons", function(x, ...) standardGeneric("exons"))

setMethod("exons", "data.frame",
    function(x, vals=NULL, columns="exon_id")
        stop("Please use 'exons_deprecated' for older ",
             "data.frame-based transcript metadata.")
)

setMethod("exons", "TranscriptDb",
    function(x, vals=NULL, columns="exon_id")
        .extractFeatureRowsAsGRanges("exon", x, vals, columns)
)

setGeneric("cds", function(x, ...) standardGeneric("cds"))

setMethod("cds", "TranscriptDb",
    function(x, vals=NULL, columns="cds_id")
        .extractFeatureRowsAsGRanges("cds", x, vals, columns)
)

